#
# Cambalache Object Data wrapper
#
# Copyright (C) 2022-2023  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject

from .cmb_base_objects import CmbBaseObjectData
from .cmb_type_info import CmbTypeDataInfo
from cambalache import getLogger, _

logger = getLogger(__name__)


class CmbObjectData(CmbBaseObjectData):
    __gsignals__ = {
        "data-added": (GObject.SignalFlags.RUN_FIRST, None, (CmbBaseObjectData,)),
        "data-removed": (GObject.SignalFlags.RUN_FIRST, None, (CmbBaseObjectData,)),
        "arg-changed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    parent = GObject.Property(type=CmbBaseObjectData, flags=GObject.ParamFlags.READWRITE)

    object = GObject.Property(type=GObject.Object, flags=GObject.ParamFlags.READWRITE)
    info = GObject.Property(type=CmbTypeDataInfo, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.args = []
        self.children = []

        super().__init__(**kwargs)

        if self.project is None:
            return

        if self.info is None:
            type_info = self.project.type_info.get(self.owner_id, None)
            if type_info:
                self.info = type_info.find_data_info(self.data_id)

        if self.object is None:
            self.object = self.project.get_object_by_id(self.ui_id, self.object_id)

        self.__populate_children()
        self.connect("notify", self._on_notify)

    def __str__(self):
        return f"CmbObjectData<{self.owner_id}:{self.info.key}> obj={self.ui_id}:{self.object_id} data={self.data_id}:{self.id}"

    def get_id_string(self):
        return f"{self.owner_id}.{self.id}"

    def get_arg(self, key):
        c = self.project.db.execute(
            "SELECT value FROM object_data_arg WHERE ui_id=? AND object_id=? AND owner_id=? AND data_id=? AND id=? AND key=?;",
            (self.ui_id, self.object_id, self.owner_id, self.data_id, self.id, key),
        )
        row = c.fetchone()
        return row[0] if row is not None else None

    def _on_notify(self, obj, pspec):
        if pspec.name in ["value"]:
            self.project._object_data_changed(self)

    def _arg_changed(self, key):
        self.emit("arg-changed", key)
        self.project._object_data_arg_changed(self, key)

    def set_arg(self, key, value):
        # Prevent potential infinite recursion
        val = self.get_arg(key)
        if val == value:
            return

        c = self.project.db.cursor()

        try:
            if value is None:
                c.execute(
                    """
                    DELETE FROM object_data_arg
                    WHERE ui_id=? AND object_id=? AND owner_id=? AND data_id=? AND id=? AND key=?;
                    """,
                    (self.ui_id, self.object_id, self.owner_id, self.data_id, self.id, key),
                )
            else:
                # Do not use REPLACE INTO, to make sure both INSERT and UPDATE triggers are used
                count = self.db_get(
                    """
                    SELECT count(value) FROM object_data_arg
                    WHERE ui_id=? AND object_id=? AND owner_id=? AND data_id=? AND id=? AND key=?;
                    """,
                    (self.ui_id, self.object_id, self.owner_id, self.data_id, self.id, key),
                )

                if count:
                    c.execute(
                        """
                        UPDATE object_data_arg SET value=?
                        WHERE ui_id=? AND object_id=? AND owner_id=? AND data_id=? AND id=? AND key=?;
                        """,
                        (str(value), self.ui_id, self.object_id, self.owner_id, self.data_id, self.id, key),
                    )
                else:
                    c.execute(
                        """
                        INSERT INTO object_data_arg (ui_id, object_id, owner_id, data_id, id, key, value)
                        VALUES (?, ?, ?, ?, ?, ?, ?);
                        """,
                        (self.ui_id, self.object_id, self.owner_id, self.data_id, self.id, key, str(value)),
                    )

            self._arg_changed(key)
        except Exception as e:
            logger.warning(f"{self} Error setting arg {key}={value}: {e}")

        c.close()

    def __add_child(self, child):
        if child in self.children:
            return

        self.children.append(child)
        self.object.data_dict[child.get_id_string()] = child
        self.emit("data-added", child)
        self.project._object_data_data_added(self, child)

    def _remove_child(self, child):
        self.children.remove(child)
        del self.object.data_dict[child.get_id_string()]
        self.emit("data-removed", child)
        self.project._object_data_data_removed(self, child)

    def _add_child(self, owner_id, data_id, id, info=None):
        new_data = CmbObjectData(
            project=self.project,
            object=self.object,
            ui_id=self.ui_id,
            object_id=self.object_id,
            owner_id=owner_id,
            data_id=data_id,
            id=id,
            parent=self,
            info=info,
        )
        self.__add_child(new_data)
        return new_data

    def __populate_children(self):
        c = self.project.db.cursor()

        # Populate children
        for row in c.execute(
            "SELECT * FROM object_data WHERE ui_id=? AND object_id=? AND owner_id=? AND parent_id=?;",
            (self.ui_id, self.object_id, self.owner_id, self.id),
        ):
            obj = CmbObjectData.from_row(self.project, *row)
            obj.parent = self
            self.__add_child(obj)

    def add_data(self, data_key, value=None, comment=None):
        try:
            value = str(value) if value is not None else None
            taginfo = self.info.children.get(data_key)
            owner_id = taginfo.owner_id
            data_id = taginfo.data_id
            id = self.project.db.object_add_data(self.ui_id, self.object_id, owner_id, data_id, value, self.id, comment)
        except Exception as e:
            logger.warning(f"{self} Error adding child data {data_key}: {e}")
            return None
        else:
            return self._add_child(owner_id, data_id, id, taginfo)

    def remove_data(self, data):
        try:
            assert data in self.children

            self.project.history_push(
                _("Remove {key} from {name}").format(key=data.info.key, name=self.object.display_name_type)
            )

            self.project.db.execute(
                "DELETE FROM object_data WHERE ui_id=? AND object_id=? AND owner_id=? AND data_id=? AND id=?;",
                (self.ui_id, self.object_id, data.owner_id, data.data_id, data.id),
            )
            self.project.db.commit()
            self.project.history_pop()
        except Exception as e:
            logger.warning(f"{self} Error removing data {data}: {e}")
            return False
        else:
            self._remove_child(data)
            return True



# GtkPopover Controller
#
# Copyright (C) 2022  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk, CambalachePrivate

from .mrg_selection import MrgSelection
from .mrg_gtk_widget import MrgGtkWidget


class MrgGtkPopover(MrgGtkWidget):
    object = GObject.Property(type=Gtk.Popover, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.window = None
        self.__button = None

        super().__init__(**kwargs)

        # Create wrapper window
        self.__button = Gtk.MenuButton(
            visible=True, halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER, receives_default=False
        )

        self.window = Gtk.Window(title="Popover Preview Window", deletable=False)
        self.window.set_default_size(320, 240)

        if Gtk.MAJOR_VERSION == 4:
            self.__button.set_icon_name("open-menu-symbolic")
            self.window.set_child(self.__button)
            self.property_ignore_list.add("autohide")
        else:
            self.__button.add(Gtk.Image(visible=True, icon_name="open-menu-symbolic"))
            self.window.add(self.__button)
            self.property_ignore_list.add("modal")

    def object_changed(self, old, new):
        self.selection = None

        if self.object is None:
            self.__button.set_popover(None)
            self.window.hide()
            return

        self.selection = MrgSelection(app=self.app, container=self.object)
        CambalachePrivate.widget_set_application_id(self.window, f"Cmb:{self.ui_id}.{self.object_id}")

        # TODO: keep track when these prop changes and update window
        if Gtk.MAJOR_VERSION == 4:
            unused = self.object.props.parent is None
        else:
            unused = self.object.props.relative_to is None

        if unused:
            self.__button.set_popover(self.object)

            if Gtk.MAJOR_VERSION == 3:
                self.window.show_all()
                self.object.set_modal(False)
            else:
                self.object.set_autohide(False)

            self.window.present()
        else:
            self.window.hide()

    def on_selected_changed(self):
        super().on_selected_changed()

        if self.__button:
            self.__button.popup()

    def show_child(self, child):
        if self.__button:
            self.__button.popup()



# GtkWindow Controller
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk, CambalachePrivate

from .mrg_gtk_bin import MrgGtkBin
from .mrg_selection import MrgSelection

from merengue import getLogger

logger = getLogger(__name__)


class MrgGtkWindow(MrgGtkBin):
    object = GObject.Property(type=Gtk.Window, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.selection = None

        super().__init__(**kwargs)

        self.selection = MrgSelection(app=self.app, container=self.object)
        self.property_ignore_list.add("modal")

    def object_changed(self, old, new):
        super().object_changed(old, new)

        if old:
            old.destroy()

        # Handle widget selection
        if self.selection:
            self.selection.container = self.object

        if self.object:
            self._update_name()

            # Make sure the user can not close the window
            if Gtk.MAJOR_VERSION == 4:
                self.object.connect("close-request", lambda o: True)
            else:
                self.object.connect("delete-event", lambda o, e: True)

            # Disable modal at runtime
            self.object.props.modal = False

            # Always show toplevels windows
            if Gtk.MAJOR_VERSION == 4:
                self.object.show()
            else:
                self.object.show_all()

            CambalachePrivate.widget_set_application_id(self.object, f"Casilda:{self.ui_id}.{self.object_id}")

            # Add gtk version CSS class
            if Gtk.MAJOR_VERSION == 4:
                self.object.add_css_class("gtk4")
            else:
                self.object.get_style_context().add_class("gtk3")

    def _update_name(self):
        if self.object is None:
            return

        # TODO: find a way to get object name instead of ID
        type_name = GObject.type_name(self.object.__gtype__)
        self.object.props.title = type_name



#
# CmbTranslatablePopover - Cambalache Translatable Popover
#
# Copyright (C) 2021  Philipp Unger
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Philipp Unger <philipp.unger.1988@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import Gtk

from .cmb_translatable_widget import CmbTranslatableWidget
from cambalache import _


class CmbTranslatablePopover(Gtk.Popover):
    __gtype_name__ = "CmbTranslatablePopover"

    def __init__(self, **kwargs):
        self._object = None
        super().__init__(**kwargs)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.append(Gtk.Label(label=_("<b>Translation</b>"), use_markup=True))
        box.append(Gtk.Separator())

        self._translation = CmbTranslatableWidget()
        box.append(self._translation)

        self.set_child(box)

    def bind_properties(self, target):
        self._translation.bind_properties(target)



#!/usr/bin/pytest

"""
Test Undo/Redo API
"""
import os

from cambalache import CmbProject


def undo_test(target_tk, filename, name):
    path = os.path.join(os.path.dirname(__file__), target_tk, filename)

    project = CmbProject(target_tk=target_tk)

    ui, msgs, detail_msg = project.import_file(path)

    for i in range(0, 4):
        obj = project.get_object_by_name(ui.ui_id, name)
        assert obj is not None and obj.name == name

        project.remove_object(obj)

        assert project.get_object_by_name(ui.ui_id, name) is None

        project.undo()

        obj = project.get_object_by_name(ui.ui_id, name)
        assert obj is not None and obj.name == name


def test_gtk3_undo():
    undo_test("gtk+-3.0", "dialog.ui", "dialog")


def test_gtk4_undo():
    undo_test("gtk-4.0", "liststore.ui", "liststore_test")



#
# CmbObjectListEditor
#
# Copyright (C) 2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk
from ..cmb_object import CmbObject


class CmbObjectListEditor(Gtk.ScrolledWindow):
    __gtype_name__ = "CmbObjectListEditor"

    parent = GObject.Property(type=CmbObject, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)
    type_id = GObject.Property(type=str, default=None, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.__updating = False
        self.__object_ids = None

        self.props.height_request = 96
        self.buffer = Gtk.TextBuffer()
        self.view = Gtk.TextView(visible=True, buffer=self.buffer)
        self.set_child(self.view)

        self.buffer.connect("notify::text", self.__on_text_notify)

    def __on_text_notify(self, obj, pspec):
        if self.__updating:
            return

        text = self.buffer.props.text
        objects = []

        if text and len(text):
            names = [name.strip() for name in text.split("\n")]
            ui_id = self.parent.ui_id
            project = self.parent.project

            for name in names:
                obj = project.get_object_by_name(ui_id, name)
                if obj is not None:
                    objects.append(str(obj.object_id))

        objects = ",".join(objects)

        if self.__object_ids == objects:
            return

        self.__object_ids = objects
        self.notify("cmb-value")

    @GObject.Property(type=str)
    def cmb_value(self):
        return self.__object_ids

    @cmb_value.setter
    def _set_cmb_value(self, value):
        value = value if value else None

        if self.__object_ids == value:
            return

        self.__object_ids = value

        self.__updating = True
        if self.__object_ids:
            names = self.parent.project._get_object_list_names(self.parent.ui_id, self.__object_ids)
            self.buffer.props.text = "\n".join(names)
        else:
            self.buffer.props.text = ""
        self.__updating = False


Gtk.WidgetClass.set_css_name(CmbObjectListEditor, "CmbObjectListEditor")



#
# CmbSwitch
#
# Copyright (C) 2021-2023  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk


class CmbSwitch(Gtk.Switch):
    __gtype_name__ = "CmbSwitch"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.connect("notify::active", self.__on_notify)
        self.props.halign = Gtk.Align.START

    def __on_notify(self, obj, pspec):
        self.notify("cmb-value")

    @GObject.Property(type=str)
    def cmb_value(self):
        return "True" if self.props.active else "False"

    @cmb_value.setter
    def _set_cmb_value(self, value):
        if value is not None:
            if type(value) is str:
                val = value.lower()
                if val in {"1", "t", "y", "true", "yes"}:
                    active = True
                else:
                    active = False
            else:
                active = bool(value)
        else:
            active = False

        if active == self.props.active:
            return

        self.props.active = active



#
# CmbSuggestionEntry
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk
from .cmb_entry import CmbEntry


class CmbSuggestionEntry(CmbEntry):
    __gtype_name__ = "CmbSuggestionEntry"

    def __init__(self, **kwargs):
        self._filters = {}

        super().__init__(**kwargs)

        # GObject.Object.bind_property(self, "cmb-value", self.props.buffer, "text", GObject.BindingFlags.SYNC_CREATE)

        self.type_model = Gtk.ListStore(str)

        # Completion
        self.__completion = Gtk.EntryCompletion()
        self.__completion.props.model = self.type_model
        self.__completion.props.text_column = 0
        self.__completion.props.inline_completion = True
        self.__completion.props.inline_selection = True
        self.__completion.props.popup_set_width = True
        self.props.completion = self.__completion

        # self.__completion.set_match_func(lambda o, key, iter, d: key in self.type_model[iter][0], None)

        # String
        renderer_text = Gtk.CellRendererText()
        self.__completion.pack_start(renderer_text, False)
        self.__completion.add_attribute(renderer_text, "text", 0)

    def set_suggestions(self, suggestions):
        self.type_model.clear()
        for suggestion in suggestions:
            self.type_model.append([suggestion])

    def __model_filter_func(self, model, iter, data):
        return model[iter][0] == data



# GtkGrid Controller
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .mrg_gtk_widget import MrgGtkWidget
from merengue import MrgPlaceholder, getLogger

logger = getLogger(__name__)


class MrgGtkGrid(MrgGtkWidget):
    object = GObject.Property(type=Gtk.Grid, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if Gtk.MAJOR_VERSION == 4:
            self._packing = ["column", "row", "column-span", "row-span"]
        else:
            self._packing = ["left-attach", "top-attach", "width", "height"]

        self.width = None
        self.height = None

    def __get_size(self, ignore_placeholders=False):
        w = 0
        h = 0

        for child in self.get_children():
            if ignore_placeholders and isinstance(child, MrgPlaceholder):
                continue
            x, y, width, height = self.child_get(child, self._packing)
            w = max(x + width, w)
            h = max(y + height, h)

        return (w, h)

    def __cleanup_placeholders(self):
        # Remove all placeholders
        children = self.get_children()
        for child in children:
            if isinstance(child, MrgPlaceholder):
                self.object.remove(child)

    def __ensure_placeholders(self, width, height):
        if self.object is None:
            return

        self.__cleanup_placeholders()

        for x in range(width):
            for y in range(height):
                if self.object.get_child_at(x, y) is None:
                    placeholder = MrgPlaceholder(visible=True, controller=self)
                    self.object.attach(placeholder, x, y, 1, 1)

        # Save size for rebuild
        self.width = width
        self.height = height

    def object_changed(self, old, new):
        super().object_changed(old, new)
        width, height = self.__get_size(ignore_placeholders=True)

        if self.width is None:
            self.width = width if width else 3
            self.height = height if width else 3

        self.__ensure_placeholders(max(width, self.width), max(height, self.height))

    def get_child_layout(self, child, layout):
        for prop in self._packing:
            layout[prop] = self.child_get(child, [prop])[0]

        return super().get_child_layout(child, layout)

    def add_placeholder(self, mod):
        width, height = self.__get_size()

        if mod:
            height += 1
        else:
            width += 1

        self.__ensure_placeholders(max(width, 1), max(height, 1))

    def remove_placeholder(self, mod):
        width, height = self.__get_size()
        width_np, height_np = self.__get_size(ignore_placeholders=True)

        if mod:
            height = max(height - 1, height_np)
        else:
            width = max(width - 1, width_np)

        self.__ensure_placeholders(max(width, 1), max(height, 1))



# MrgSelection Selection handling
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from merengue import utils, MrgPlaceholder


class FindInContainerData:
    def __init__(self, toplevel, x, y):
        self.toplevel = toplevel
        self.x = x
        self.y = y
        self.child = None
        self.level = None


class MrgSelection(GObject.GObject):
    app = GObject.Property(type=GObject.GObject, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.__selecting = False
        self._container = None
        self.gesture = None

        super().__init__(**kwargs)

    @GObject.property(type=Gtk.Widget)
    def container(self):
        return self._container

    @container.setter
    def _set_container(self, obj):
        self._container = obj

        if self._container:
            self.gesture = utils.gesture_click_new(self._container, propagation_phase=Gtk.PropagationPhase.CAPTURE)
            self.gesture.connect("pressed", self.__on_gesture_button_pressed)
            self.gesture.connect("released", self.__on_gesture_button_released)
        elif self.gesture:
            self.gesture.disconnect_by_func(self.__on_gesture_button_pressed)
            self.gesture.disconnect_by_func(self.__on_gesture_button_released)
            self.gesture = None

    def __on_gesture_button_pressed(self, gesture, n_press, x, y):
        child = self.get_child_at_position(self._container, x, y)

        if isinstance(child, MrgPlaceholder):
            controller = child.controller

            # Write placeholder selected/activated message
            if n_press == 2:
                child.activated()
            else:
                child.selected()
        else:
            controller = self.app.get_controller_from_object(child)

        if controller is None or controller.selected:
            return

        object_id = utils.object_get_id(controller.object)
        if object_id is None:
            return

        # Select widget on button release only if its preselected
        self.app.write_command("selection_changed", args={"selection": [object_id]})
        controller.selected = True

        if not isinstance(child, Gtk.Window) and not isinstance(child, Gtk.HeaderBar):
            gesture.set_state(Gtk.EventSequenceState.CLAIMED)
            self.__selecting = True

    def __on_gesture_button_released(self, gesture, n_press, x, y):
        if self.__selecting:
            gesture.set_state(Gtk.EventSequenceState.CLAIMED)
            self.__selecting = False

    def is_widget_from_ui(self, obj):
        if isinstance(obj, MrgPlaceholder):
            return True

        object_id = utils.object_get_builder_id(obj)
        return object_id is not None and object_id.startswith("__cmb__")

    def _find_first_child_inside_container(self, widget, data):
        if data.child is not None or not widget.get_mapped():
            return

        x, y = data.toplevel.translate_coordinates(widget, data.x, data.y)

        w = widget.get_allocated_width()
        h = widget.get_allocated_height()

        if x >= 0 and x < w and y >= 0 and y < h:
            from_ui = self.is_widget_from_ui(widget)

            if issubclass(type(widget), Gtk.Container):
                if from_ui:
                    data.child = self.get_child_at_position(widget, x, y)
                else:
                    widget.forall(self._find_first_child_inside_container, data)

            if data.child is None and from_ui:
                data.child = widget

    def get_child_at_position(self, widget, x, y):
        if Gtk.MAJOR_VERSION == 4:
            pick = widget.pick(x, y, Gtk.PickFlags.INSENSITIVE | Gtk.PickFlags.NON_TARGETABLE)
            while pick and not self.is_widget_from_ui(pick):
                pick = pick.props.parent
            return pick

        if not widget.get_mapped():
            return None

        w = widget.get_allocated_width()
        h = widget.get_allocated_height()

        if x >= 0 and x <= w and y >= 0 and y <= h:
            if issubclass(type(widget), Gtk.Container):
                data = FindInContainerData(widget, x, y)

                widget.forall(self._find_first_child_inside_container, data)

                return data.child if data.child is not None else widget
            else:
                return widget

        return None




# Cambalache
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
import gi
import logging
import gettext
import builtins

from . import config

gi.require_version("GIRepository", "3.0")
gi.require_version("Gdk", "4.0")
gi.require_version("Gtk", "4.0")
gi.require_version('Casilda', '1.0')
gi.require_version("GtkSource", "5")
gi.require_version("WebKit", "6.0")
gi.require_version('Adw', '1')

# flake8: noqa: E402,F401
from gi.repository import Gio, Gdk, Gtk

# Ensure _() builtin
if "_" not in builtins.__dict__:
    _ = gettext.gettext

if "ngettext" not in builtins.__dict__:
    ngettext = gettext.ngettext

resource = Gio.Resource.load(os.path.join(config.pkgdatadir, "cambalache.gresource"))
resource._register()

provider = Gtk.CssProvider()
provider.load_from_resource("/ar/xjuan/Cambalache/cambalache.css")
display = Gdk.Display.get_default()
Gtk.StyleContext.add_provider_for_display(display, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)


# FIXME: this is needed in flatpak for icons to work
Gtk.IconTheme.get_for_display(display).add_search_path("/app/share/icons")


def getLogger(name):
    formatter = logging.Formatter("%(levelname)s:%(name)s %(message)s")

    ch = logging.StreamHandler()
    ch.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(os.environ.get("CAMBALACHE_LOGLEVEL", "WARNING").upper())
    logger.addHandler(ch)

    return logger

from .cmb_base_objects import CmbBaseObject
from .cmb_base_file_monitor import CmbBaseFileMonitor
from .cmb_css import CmbCSS
from .cmb_ui import CmbUI
from .cmb_object import CmbObject
from .cmb_gresource import CmbGResource

# from .cmb_object_data import CmbObjectData
from .cmb_property import CmbProperty
from .cmb_property_label import CmbPropertyLabel
from .cmb_layout_property import CmbLayoutProperty
from .cmb_type_info import CmbTypeInfo
from .cmb_project import CmbProject

from .cmb_db_inspector import CmbDBInspector
from .cmb_file_status_bar import CmbFileStatusBar
from .cmb_graphics_offload import CmbGraphicsOffload
from .cmb_view import CmbView
from .cmb_list_view import CmbListView
from .cmb_notification import notification_center, CmbNotification, CmbNotificationCenter
from .cmb_notification_list_view import CmbNotificationListView
from .cmb_object_property_editor import CmbObjectPropertyEditor
from .cmb_object_editor import CmbObjectEditor
from .cmb_signal_editor import CmbSignalEditor
from .cmb_ui_property_editor import CmbUIPropertyEditor
from .cmb_ui_requires_editor import CmbUIRequiresEditor
from .cmb_ui_editor import CmbUIEditor
from .cmb_css_editor import CmbCSSEditor
from .cmb_gresource_editor import CmbGResourceEditor
from .cmb_fragment_editor import CmbFragmentEditor
from .cmb_accessible_editor import CmbAccessibleEditor
from .cmb_type_chooser import CmbTypeChooser
from .cmb_type_chooser_widget import CmbTypeChooserWidget
from .cmb_type_chooser_popover import CmbTypeChooserPopover
from .cmb_project_settings import CmbProjectSettings



# GtkListBox Controller
#
# Copyright (C) 2022  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .mrg_gtk_box import MrgGtkBox
from merengue import MrgPlaceholder, utils


class MrgGtkListBox(MrgGtkBox):
    object = GObject.Property(type=Gtk.ListBox, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_children(self):
        children = super().get_children()
        retval = []

        if children is None:
            return None

        for child in children:
            id = utils.object_get_id(child)
            if id:
                retval.append(child)
            else:
                grandchild = child.props.child if Gtk.MAJOR_VERSION == 4 else child.get_child()
                id = utils.object_get_id(grandchild)
                if id or isinstance(grandchild, MrgPlaceholder):
                    retval.append(grandchild)

        return retval

    def get_child_layout(self, child, layout):
        # GtkListBox has not layout properties
        return layout

    def remove_child(self, child):
        if self.object is None:
            return

        if isinstance(child, Gtk.ListBoxRow):
            super().remove_child(child)
        else:
            super().remove_child(child.props.parent)



#
# CmbNotificationListRow
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import datetime

from cambalache import getLogger
from gi.repository import GObject, Gtk

logger = getLogger(__name__)


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_notification_list_row.ui")
class CmbNotificationListRow(Gtk.ListBoxRow):
    __gtype_name__ = "CmbNotificationListRow"

    view = GObject.Property(type=Gtk.Widget, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    box = Gtk.Template.Child()
    date_label = Gtk.Template.Child()
    close_button = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.props.activatable = False

        notification = self.view.notification
        start_date = datetime.datetime.utcfromtimestamp(notification.start_date).strftime("%x")
        self.date_label.set_label(f"<small>{start_date}</small>")
        self.box.prepend(self.view)

        self.add_css_class("cmb-notification-list-row")

    @Gtk.Template.Callback("on_map")
    def __on_map(self, w):
        self.props.child.props.reveal_child = True

    @Gtk.Template.Callback("on_close_button_clicked")
    def __on_close_button_clicked(self, button):
        notification = self.view.notification
        notification.center.remove(notification)



# Merengue Gtk plugin
#
# Copyright (C) 2021-2022  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .mrg_gtk_aspect_frame import MrgGtkAspectFrame
from .mrg_gtk_assistant import MrgGtkAssistant
from .mrg_gtk_bin import MrgGtkBin

if Gtk.MAJOR_VERSION == 3:
    from .mrg_gtk_menu_item import MrgGtkMenuItem
    from .mrg_gtk_menu import MrgGtkMenu
elif Gtk.MAJOR_VERSION == 4:
    from .mrg_gtk_center_box import MrgGtkCenterBox
    from .mrg_gtk_stack_page import MrgGtkStackPage

from .mrg_gtk_box import MrgGtkBox
from .mrg_gtk_dialog import MrgGtkDialog
from .mrg_gtk_expander import MrgGtkExpander
from .mrg_gtk_frame import MrgGtkFrame
from .mrg_gtk_grid import MrgGtkGrid
from .mrg_gtk_label import MrgGtkLabel
from .mrg_gtk_list_box import MrgGtkListBox
from .mrg_gtk_list_box_row import MrgGtkListBoxRow
from .mrg_gtk_notebook import MrgGtkNotebook
from .mrg_gtk_overlay import MrgGtkOverlay
from .mrg_gtk_paned import MrgGtkPaned
from .mrg_gtk_popover import MrgGtkPopover
from .mrg_gtk_revealer import MrgGtkRevealer
from .mrg_gtk_scrolled_window import MrgGtkScrolledWindow
from .mrg_gtk_stack import MrgGtkStack
from .mrg_gtk_viewport import MrgGtkViewport
from .mrg_gtk_widget import MrgGtkWidget
from .mrg_gtk_window import MrgGtkWindow
from .mrg_selection import MrgSelection

# Proxy objects
# Some objects can not be instantiated in the runtime if not properly configured
from .mrg_g_binding import MrgGBindingProxy

GObject.type_ensure(MrgGBindingProxy.__gtype__)




#
# Cambalache Library Info wrapper
#
# Copyright (C) 2022-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject
from .cmb_base_objects import CmbBaseLibraryInfo
from .cmb_type_info import CmbTypeInfo
from cambalache import _, CmbObject


class CmbLibraryInfo(CmbBaseLibraryInfo):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Type Information
        self.type_info = {}

        self.object_types = self.__init_object_types()
        self.min_version = self.__init_min_version()

        # Init type_info for this library
        for row in self.project.db.execute(
            "SELECT * FROM type WHERE parent_id IS NOT NULL AND library_id=? ORDER BY type_id;",
            (self.library_id, )
        ):
            type_id = row[0]
            info = CmbTypeInfo.from_row(self.project, *row)
            self.type_info[type_id] = info

        # Set parent back reference
        for type_id, info in self.type_info.items():
            info.parent = self.type_info.get(info.parent_id, None) or self.project.type_info.get(info.parent_id)

    def __str__(self):
        return f"CmbLibraryInfo<{self.library_id}-{self.version}>"

    def __init_object_types(self):
        prefix = self.prefix
        prefix_len = len(prefix)
        retval = []

        for row in self.project.db.execute("SELECT type_id FROM type WHERE library_id=?", (self.library_id,)):
            (type_id,) = row
            if type_id.startswith(prefix):
                # Remove Prefix from type name
                retval.append(type_id[prefix_len:])

        return retval

    def __init_min_version(self):
        row = self.project.db.execute(
            "SELECT MIN_VERSION(version) FROM library_version WHERE library_id=?;", (self.library_id,)
        ).fetchone()

        return row[0] if row is not None else None

    @GObject.Property(type=bool, default=False)
    def enabled(self):
        return self.db_get("SELECT enabled FROM library WHERE (library_id) IS (?);", (self.library_id,))

    @enabled.setter
    def _set_enabled(self, value):
        if self.enabled == value:
            return

        self.db_set("UPDATE library SET enabled=? WHERE (library_id) IS (?);", (self.library_id,), value)

        # Load or Unload type infos from project
        if value:
            self.project.type_info.update(self.type_info)
            for type_id, info in self.type_info.items():
                self.project.emit("type-info-added", info)
        else:
            for type_id in self.type_info.keys():
                info = self.project.type_info.pop(type_id)
                self.project.emit("type-info-removed", info)

        self.project._library_info_changed(self, "enabled")




#
# Blueprint compiler integration functions
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import io

try:
    import blueprintcompiler as bp
    from blueprintcompiler import parser, tokenizer
    from blueprintcompiler.decompiler import decompile_string
    from blueprintcompiler.outputs import XmlOutput
except Exception:
    bp = None


class CmbBlueprintError(Exception):
    def __init__(self, message, errors=[]):
        super().__init__(message)
        self.errors = errors


class CmbBlueprintUnsupportedError(CmbBlueprintError):
    pass


class CmbBlueprintMissingError(CmbBlueprintError):
    def __init__(self):
        super().__init__("blueprintcompiler is not available")


def cmb_blueprint_decompile(data: str) -> str:
    if bp is None:
        raise CmbBlueprintMissingError()

    try:
        retval = decompile_string(data)
    except bp.decompiler.UnsupportedError as e:
        raise CmbBlueprintUnsupportedError(str(e))
    except Exception as e:
        raise CmbBlueprintError(str(e))

    return retval


def cmb_blueprint_compile(data: str) -> str:
    if bp is None:
        raise CmbBlueprintMissingError()

    tokens = tokenizer.tokenize(data)
    ast, errors, warnings = parser.parse(tokens)

    if errors:
        f = io.StringIO("")
        errors.pretty_print("temp", data, f)
        f.seek(0)
        raise CmbBlueprintError(f.read(), errors=errors)

    if ast is None:
        raise CmbBlueprintError("AST is None")

    # Ignore warnings

    retval = XmlOutput().emit(ast)
    return retval.encode()




#
# CmbObjectChooser
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from cambalache import _
from gi.repository import GObject, Gdk, Gtk

from ..cmb_object import CmbObject
from ..cmb_type_chooser_popover import CmbTypeChooserPopover


class CmbObjectChooser(Gtk.Entry):
    __gtype_name__ = "CmbObjectChooser"

    parent = GObject.Property(type=CmbObject, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)
    is_inline = GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)
    inline_object_id = GObject.Property(type=str, default=None, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)
    inline_property_id = GObject.Property(type=str, default=None, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)
    type_id = GObject.Property(type=str, default="GObject", flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)

    def __init__(self, **kwargs):
        self.__object_id = None
        self.__updating = None
        super().__init__(**kwargs)

        self.connect("notify::parent", self.__on_parent_notify)
        self.connect("notify::text", self.__on_text_notify)

        if self.is_inline:
            self.connect("icon-press", self.__on_icon_pressed)
            if self.parent:
                self.parent.connect("property-changed", self.__on_parent_property_changed)
            self.__update_icons()
        else:
            self.props.placeholder_text = f"<{self.type_id}>"

        drop_target = Gtk.DropTarget.new(
            type=GObject.TYPE_NONE, actions=Gdk.DragAction.COPY
        )
        drop_target.set_gtypes([CmbObject])
        drop_target.connect("accept", self.__on_drop_accept)
        drop_target.connect("drop", self.__on_drop_drop)
        self.add_controller(drop_target)

    def __on_parent_notify(self, obj, pspec):
        if self.parent:
            self.parent.connect("property-changed", self.__on_parent_property_changed)

    def __on_text_notify(self, obj, pspec):
        if self.inline_object_id:
            return

        if self.__updating:
            self.__updating = False
            return

        obj = self.parent.project.get_object_by_name(self.parent.ui_id, self.props.text)
        value = obj.object_id if obj else None

        if self.__object_id != value:
            self.__object_id = value
            self.notify("cmb-value")

    def __on_parent_property_changed(self, parent, prop, field):
        if not self.is_inline or prop.property_id != self.inline_property_id:
            return

        self.inline_object_id = prop.inline_object_id
        self.__update_icons()

    def __on_drop_accept(self, target, drop):
        drag = drop.get_drag()
        origin_item = drag._item

        if origin_item == self.parent:
            return False

        return origin_item.info.is_a(self.type_id)

    def __on_drop_drop(self, target, origin_item, x, y):
        if self.is_inline:
            project = self.parent.project
            prop = self.parent.properties_dict.get(self.inline_property_id, None)

            if prop is None:
                return

            project.history_push(_("Move {name} to {property}").format(
                name=origin_item.display_name,
                property=f"{self.parent.type_id}::{self.inline_property_id}")
            )

            if prop.inline_object_id:
                old_inline_object = project.get_object_by_id(prop.ui_id, prop.inline_object_id)
                project.remove_object(old_inline_object)

            # TODO: implement this in plain SQL
            parent_id = self.parent.object_id
            object_id = origin_item.object_id

            origin_item.parent_id = parent_id
            prop.value = object_id
            prop.inline_object_id = object_id
            self.inline_object_id = object_id
            self.__update_icons()

            project.history_pop()
        else:
            # TODO: ensure dragged object has an id
            # Select dragged object id
            self.cmb_value = origin_item.object_id

    @GObject.Property(type=str)
    def cmb_value(self):
        return str(self.__object_id) if self.__object_id else None

    @cmb_value.setter
    def _set_cmb_value(self, value):
        parent = self.parent
        value = value if value else None

        if self.__object_id == value:
            return

        self.__object_id = value

        self.__updating = True
        if self.__object_id:
            obj = parent.project.get_object_by_id(parent.ui_id, self.__object_id)
            self.props.text = obj.name if obj else ""
        else:
            self.props.text = ""

    def __update_icons(self):
        if not self.is_inline:
            return

        if self.inline_object_id:
            obj = self.parent.project.get_object_by_id(self.parent.ui_id, self.inline_object_id)
            type = obj.type_id
            self.props.secondary_icon_name = "edit-clear-symbolic"
            self.props.secondary_icon_tooltip_text = _("Clear property")
            self.props.placeholder_text = f"<inline {type}>"
            self.props.editable = False
            self.props.can_focus = False
        else:
            self.props.secondary_icon_name = "list-add-symbolic"
            self.props.secondary_icon_tooltip_text = _("Add inline object")
            self.props.placeholder_text = f"<{self.type_id}>"
            self.props.editable = True
            self.props.can_focus = True

    def __get_name_for_object(self, obj):
        name = obj.name
        return obj.type_id.lower() if name is None else name

    def __on_type_selected(self, popover, info):
        parent = self.parent
        parent.project.add_object(parent.ui_id, info.type_id, parent_id=parent.object_id, inline_property=self.inline_property_id)
        self.__update_icons()

    def __on_icon_pressed(self, widget, icon_pos):
        if not self.is_inline:
            return

        parent = self.parent
        project = parent.project

        if self.inline_object_id:
            obj = project.get_object_by_id(self.parent.ui_id, self.inline_object_id)
            project.remove_object(obj)
            self.inline_object_id = None
            self.__update_icons()
        else:
            chooser = CmbTypeChooserPopover(parent_type_id=parent.type_id, derived_type_id=self.type_id)
            chooser.set_parent(self)
            chooser.project = project
            chooser.connect("type-selected", self.__on_type_selected)
            chooser.popup()



#
# CmbSpinButton
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import math

from gi.repository import GObject, GLib, Gtk


class CmbSpinButton(Gtk.SpinButton):
    __gtype_name__ = "CmbSpinButton"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.connect("notify::value", self.__on_value_notify)
        self.props.halign = Gtk.Align.START
        self.props.numeric = True
        self.props.width_chars = 8
        self.__ignore_notify = False
        self.__cmb_value = None

    def __on_value_notify(self, obj, pspec):
        if self.__ignore_notify:
            return

        self.cmb_value = self.props.value

    @GObject.Property(type=str)
    def cmb_value(self):
        return self.__cmb_value

    @cmb_value.setter
    def _set_cmb_value(self, value):
        value = float(value)

        if value == math.inf:
            value = GLib.MAXDOUBLE
        elif value == -math.inf:
            value = -GLib.MAXDOUBLE
        else:
            value = value

        if self.props.digits == 0:
            cmb_value = str(int(value))
        else:
            # NOTE: round() to avoid setting numbers like 0.7000000000000001
            cmb_value = str(round(value, 15))

        if value == cmb_value:
            return

        self.__ignore_notify = True
        self.__cmb_value = cmb_value
        self.props.value = value
        self.__ignore_notify = False



# Cambalache Property Controls
#
# Copyright (C) 2021-2023  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
import math

from gi.repository import GLib, Gtk
from .cmb_boolean_undefined import CmbBooleanUndefined
from .cmb_entry import CmbEntry
from .cmb_file_entry import CmbFileEntry
from .cmb_icon_name_entry import CmbIconNameEntry
from .cmb_pixbuf_entry import CmbPixbufEntry
from .cmb_spin_button import CmbSpinButton
from .cmb_color_entry import CmbColorEntry
from .cmb_enum_combo_box import CmbEnumComboBox
from .cmb_flags_entry import CmbFlagsEntry
from .cmb_object_chooser import CmbObjectChooser
from .cmb_object_list_editor import CmbObjectListEditor
from .cmb_switch import CmbSwitch
from .cmb_text_view import CmbTextView
from .cmb_suggestion_entry import CmbSuggestionEntry
from .cmb_translatable_popover import CmbTranslatablePopover


def cmb_create_editor(project, type_id, prop=None, data=None, parent=None):
    def get_min_max_for_type(type_id):
        if type_id == "gchar":
            return (GLib.MININT8, GLib.MAXINT8)
        elif type_id == "guchar":
            return (0, GLib.MAXUINT8)
        elif type_id == "gint":
            return (GLib.MININT, GLib.MAXINT)
        elif type_id == "guint":
            return (0, GLib.MAXUINT)
        elif type_id == "glong":
            return (GLib.MINLONG, GLib.MAXLONG)
        elif type_id == "gulong":
            return (0, GLib.MAXULONG)
        elif type_id == "gint64":
            return (GLib.MININT64, GLib.MAXINT64)
        elif type_id == "guint64":
            return (0, GLib.MAXUINT64)
        elif type_id == "gfloat":
            return (-GLib.MAXFLOAT, GLib.MAXFLOAT)
        elif type_id == "gdouble":
            return (-GLib.MAXDOUBLE, GLib.MAXDOUBLE)

    def get_dirname():
        if project.filename:
            return os.path.dirname(project.filename)
        else:
            return os.getcwd()

    def on_translatable_edit_clicked(widget, target):
        popover = CmbTranslatablePopover()
        popover.set_parent(widget)
        popover.bind_properties(target)
        popover.popup()
        popover = None

    editor = None
    info = project.type_info.get(type_id, None)

    if prop:
        translatable = prop.info.translatable
    elif data:
        translatable = data.info.translatable
    else:
        translatable = False

    if type_id == "gboolean":
        editor = CmbSwitch()
    if type_id == "gunichar":
        editor = CmbEntry(hexpand=True, max_length=1, placeholder_text=f"<{type_id}>")
    elif (
        type_id == "gchar"
        or type_id == "guchar"
        or type_id == "gint"
        or type_id == "guint"
        or type_id == "glong"
        or type_id == "gulong"
        or type_id == "gint64"
        or type_id == "guint64"
        or type_id == "gfloat"
        or type_id == "gdouble"
    ):
        digits = 0
        step_increment = 1
        minimum, maximum = get_min_max_for_type(type_id)

        pinfo = prop.info if prop else None

        # FIXME: is there a better way to handle inf -inf values other
        # than casting to str?
        if pinfo and pinfo.minimum is not None:
            value = float(minimum)
            minimum = value if value != -math.inf else -GLib.MAXDOUBLE
        if pinfo and pinfo.maximum is not None:
            value = float(maximum)
            maximum = value if value != math.inf else GLib.MAXDOUBLE

        if type_id == "gfloat" or type_id == "gdouble":
            digits = 4
            step_increment = 0.1

        adjustment = Gtk.Adjustment(lower=minimum, upper=maximum, step_increment=step_increment, page_increment=10)

        editor = CmbSpinButton(digits=digits, adjustment=adjustment)
    elif type_id == "GBytes":
        editor = CmbTextView(hexpand=True)
    elif type_id in ["GStrv", "CmbText"]:
        editor = CmbTextView(hexpand=True, translatable=translatable)
        editor.connect("edit-translatable", on_translatable_edit_clicked, prop or data)
    elif type_id == "GdkRGBA":
        editor = CmbColorEntry()
    elif type_id == "GdkColor":
        editor = CmbColorEntry(use_color=True)
    elif type_id == "GdkPixbuf":
        editor = CmbPixbufEntry(hexpand=True, dirname=get_dirname())
    elif type_id == "GFile":
        editor = CmbFileEntry(hexpand=True, dirname=get_dirname())
    elif type_id == "CmbIconName":
        editor = CmbIconNameEntry(hexpand=True, placeholder_text="<Icon Name>")
    elif type_id in ["GtkShortcutTrigger", "GtkShortcutAction"]:
        editor = CmbEntry(hexpand=True, placeholder_text=f"<{type_id}>")
    elif type_id == "CmbBooleanUndefined":
        editor = CmbBooleanUndefined()
    elif type_id == "CmbAccessibleList":
        editor = CmbObjectListEditor(
            parent=prop.object if prop else parent,
            type_id="GtkAccessible",
        )
    elif type_id == "gtype":
        editor = CmbSuggestionEntry()
        editor.set_suggestions(project._get_types())
    elif info:
        if info.is_object or info.parent_id == "interface" or type_id == "GtkExpression":
            if prop is None:
                editor = CmbObjectChooser(
                    parent=parent,
                    type_id=type_id,
                )
            else:
                editor = CmbObjectChooser(
                    parent=prop.object,
                    is_inline=project.target_tk == "gtk-4.0" and not prop.info.disable_inline_object,
                    inline_object_id=prop.inline_object_id,
                    inline_property_id=prop.property_id,
                    type_id=type_id,
                )
        elif info.parent_id == "enum":
            editor = CmbEnumComboBox(info=info)
        elif info.parent_id == "flags":
            editor = CmbFlagsEntry(info=info)

    if editor is None:
        editor = CmbEntry(hexpand=True, placeholder_text=f"<{type_id}>", translatable=translatable)
        editor.connect("edit-translatable", on_translatable_edit_clicked, prop or data)

    editor.show()

    return editor



# GtkWidget Controller
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk, CambalachePrivate

from .mrg_selection import MrgSelection

from merengue import MrgController, getLogger, utils

logger = getLogger(__name__)


class MrgGtkWidget(MrgController):
    object = GObject.Property(type=Gtk.Widget, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.window = None
        self.selection = None

        super().__init__(**kwargs)

        # Make sure all widget are always visible
        self.property_ignore_list.add("visible")

        self.child_property_ignore_list = set()

        self.connect("notify::selected", self.__on_selected_changed)

        # Make sure show_all() always works
        if Gtk.MAJOR_VERSION == 3:
            self.property_ignore_list.add("no-show-all")

    def object_changed(self, old, new):
        super().object_changed(old, new)

        def window_remove_child(window):
            if window is None:
                return

            if Gtk.MAJOR_VERSION == 4:
                window.set_child(None)
            else:
                child = window.get_child()
                if child:
                    window.remove(child)

        self.on_selected_changed()

        if self.object is None:
            window_remove_child(self.window)

            if self.window:
                self.window.hide()
            return

        if not self.toplevel or issubclass(type(self.object), Gtk.Window) or self.object.props.parent is not None:
            # Make sure widget is visible in workspace
            self.object.set_visible(True)
            return

        if self.window is None:
            self.window = Gtk.Window(deletable=False)
            self.selection = MrgSelection(app=self.app, container=self.window)

        # Update title
        self.window.set_title(GObject.type_name(self.object.__gtype__))

        window_remove_child(self.window)

        if Gtk.MAJOR_VERSION == 4:
            self.window.set_child(self.object)
        else:
            self.window.add(self.object)

        # Make sure widget is visible in workspace
        self.object.set_visible(True)

        self.window.show()

        CambalachePrivate.widget_set_application_id(self.window, f"Casilda:{self.ui_id}.{self.object_id}")

    def on_selected_changed(self):
        if self.object is None:
            return

        # Update toplevel backdrop state
        if Gtk.MAJOR_VERSION == 4:
            if self.selected:
                self.object.add_css_class("merengue_selected")
            else:
                self.object.remove_css_class("merengue_selected")

            toplevel = self.object.get_root()
        else:
            if self.selected:
                self.object.get_style_context().add_class("merengue_selected")
            else:
                self.object.get_style_context().remove_class("merengue_selected")

            toplevel = self.object.get_toplevel()

        if toplevel and self.selected:
            toplevel.present()

        # This will make the selected object shown in the workspace
        if self.selected:
            self.show()

    def __on_selected_changed(self, obj, pspec):
        self.on_selected_changed()

    def __get_children(self, obj):
        if obj is None:
            return []

        if Gtk.MAJOR_VERSION == 4:
            retval = []

            child = obj.get_first_child()
            while child is not None:
                retval.append(child)
                child = child.get_next_sibling()

            return retval
        else:
            return obj.get_children() if isinstance(obj, Gtk.Container) else []

    def get_children(self):
        return self.__get_children(self.object)

    def child_get(self, child, properties):
        if self.object is None:
            return None

        if Gtk.MAJOR_VERSION == 4:
            layout_child = None
            manager = self.object.get_layout_manager()
            if manager:
                layout_child = manager.get_layout_child(child)

            return [layout_child.get_property(x) for x in properties] if layout_child else None
        else:
            return self.object.child_get(child, *properties)

    def get_child_position(self, child):
        return -1

    def get_child_type(self, child):
        return None

    def get_child_layout(self, child, layout):
        return layout

    def remove_child(self, child):
        if self.object is None:
            return

        if Gtk.MAJOR_VERSION == 3:
            self.object.remove(child)
        else:
            logger.warning("Not implemented")

    def placeholder_selected(self, placeholder):
        position = self.get_child_position(placeholder)
        child_type = self.get_child_type(placeholder)
        layout = self.get_child_layout(placeholder, {})
        self.app.write_command(
            "placeholder_selected",
            args={
                "ui_id": self.ui_id,
                "object_id": self.object_id,
                "position": position,
                "child_type": child_type,
                "layout": layout,
            },
        )

    def placeholder_activated(self, placeholder):
        position = self.get_child_position(placeholder)
        child_type = self.get_child_type(placeholder)
        layout = self.get_child_layout(placeholder, {})
        self.app.write_command(
            "placeholder_activated",
            args={
                "ui_id": self.ui_id,
                "object_id": self.object_id,
                "position": position,
                "child_type": child_type,
                "layout": layout,
            },
        )

    def add_placeholder(self, mod):
        pass

    def remove_placeholder(self, mod):
        pass

    def find_child_property(self, child, property_id):
        if self.object is None:
            return None

        if Gtk.MAJOR_VERSION == 4:
            manager = self.object.get_layout_manager()
            layout_child = manager.get_layout_child(child)
            return layout_child.find_property(property_id) if layout_child else None

        return self.object.find_child_property(property_id)

    def set_object_child_property(self, child, property_id, val):
        if self.object is None or property_id in self.child_property_ignore_list:
            return

        if Gtk.MAJOR_VERSION == 4:
            manager = self.object.get_layout_manager()
            layout_child = manager.get_layout_child(child)
            CambalachePrivate.object_set_property_from_string(layout_child, property_id, val)
        else:
            CambalachePrivate.container_child_set_property_from_string(self.object, child, property_id, val)

    def show(self):
        child = self.object
        parent = child.props.parent
        while parent:
            parent_id = utils.object_get_id(parent)
            controller = self.app.controllers.get(parent_id, None)

            if controller:
                controller.show_child(child)
                child = parent

            parent = parent.props.parent

    def show_child(self, child):
        pass

    def find_child(self, *args):
        retval = self.object

        for name in args:
            for child in self.__get_children(retval):
                if utils.object_get_builder_id(child) == name:
                    retval = child
                    break

            if utils.object_get_builder_id(retval) != name:
                return None

        return retval



#
# CmbUIRequiresEditor - Cambalache UI Requires Editor
#
# Copyright (C) 2023-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .cmb_ui import CmbUI
from . import utils


class CmbUIRequiresEditor(Gtk.Grid):
    __gtype_name__ = "CmbUIRequiresEditor"

    def __init__(self, **kwargs):
        self.__object = None
        self.__combos = {}

        super().__init__(**kwargs)

        self.props.column_spacing = 4
        self.props.row_spacing = 4

    @GObject.Property(type=CmbUI)
    def object(self):
        return self.__object

    @object.setter
    def _set__object(self, obj):
        if obj == self.__object:
            return

        if self.__object:
            self.__object.project.disconnect_by_func(self.__on_project_added_removed)
            self.__object.disconnect_by_func(self.__on_library_changed)

        self.__object = obj
        self.set_sensitive(obj is not None)

        if obj:
            self.__object.project.connect("object-added", self.__on_project_added_removed)
            self.__object.project.connect("object-removed", self.__on_project_added_removed)
            self.__object.connect("library-changed", self.__on_library_changed)
            self.__update()

    def __on_project_added_removed(self, project, obj):
        self.__update()

    def __on_library_changed(self, ui, lib):
        combo = self.__combos.get(lib, None)
        if combo:
            combo.set_active_id(ui.get_library(lib))

    def __update(self):
        self.__combos = {}

        for child in utils.widget_get_children(self):
            self.remove(child)

        if self.__object is None:
            return

        i = 0
        for library_id, data in self.__object.list_libraries().items():
            label = Gtk.Label(label=library_id, visible=True, halign=Gtk.Align.START)
            combo = self.__combobox_new(library_id, **data)

            # Keep a reference by library_id
            self.__combos[library_id] = combo

            # Append to grid
            self.attach(label, 1, i, 1, 1)
            self.attach(combo, 2, i, 1, 1)
            i += 1

    def __on_combobox_changed(self, combo, library_id):
        if self.__object:
            self.__object.set_library(library_id, combo.get_active_id())

    def __combobox_new(self, library_id, versions=[], target=None):
        combo = Gtk.ComboBoxText(visible=True)

        combo.append(None, "")

        for version in versions:
            combo.append(version, version)

        if target:
            combo.set_active_id(target)

        combo.connect("changed", self.__on_combobox_changed, library_id)

        return combo



#
# CmbPropertyLabel
#
# Copyright (C) 2023-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk, Gio, GLib

from .cmb_property import CmbProperty
from .cmb_layout_property import CmbLayoutProperty
from .cmb_binding_popover import CmbBindingPopover


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_property_label.ui")
class CmbPropertyLabel(Gtk.Button):
    __gtype_name__ = "CmbPropertyLabel"

    prop = GObject.Property(type=CmbProperty, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)
    layout_prop = GObject.Property(
        type=CmbLayoutProperty, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY
    )
    bindable = GObject.Property(type=bool, default=True, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    label = Gtk.Template.Child()
    bind_icon = Gtk.Template.Child()
    reset_button = Gtk.Template.Child()
    serialize_check = Gtk.Template.Child()
    menu = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if not self.prop and not self.layout_prop:
            raise Exception("CmbPropertyLabel requires prop or layout_prop to be set")
            return

        self.__on_init = True

        # Update label status
        if self.prop:
            # A11y properties are prefixed to avoid clashes, do not show prefix here
            self.label.props.label = self.prop.info.a11y_property_id if self.prop.info.is_a11y else self.prop.property_id

            self.__update_property_label()
            self.prop.connect("notify", self.__on_notify)

            if self.bindable:
                self.connect("clicked", self.__on_bind_button_clicked)

            self.reset_button.set_sensitive(self.prop.has_value())
            self.reset_button.props.child.props.halign = Gtk.Align.START

            self.serialize_check.props.active = self.prop.serialize_default_value
        elif self.layout_prop:
            self.bind_icon.props.visible = False
            self.label.props.label = self.layout_prop.property_id
            self.__update_layout_property_label()
            self.layout_prop.connect("notify::value", lambda o, p: self.__update_layout_property_label())

            # TODO add support for layout property reset

        # Context menu
        self.menu.set_parent(self)
        self.__click_gesture = Gtk.GestureClick(propagation_phase=Gtk.PropagationPhase.CAPTURE, button=3)
        self.__click_gesture.connect("released", self.__on_click_gesture_released)
        self.add_controller(self.__click_gesture)

        self.__on_init = False

    def __on_notify(self, prop, pspec):
        if pspec.name in {
            "value",
            "inline-object-id",
            "bind-source-id",
            "bind-owner-id",
            "bind-property-id",
            "bind-flags",
            "binding-expression-id",
            "binding-expression-object-id"
        }:
            self.__update_property_label()

    def __update_label(self, prop):
        if prop.value != prop.info.default_value:
            self.add_css_class("modified")
        else:
            self.remove_css_class("modified")

        msg = prop.version_warning
        self.set_tooltip_markup(msg)

        if msg:
            self.add_css_class("warning")
        else:
            self.remove_css_class("warning")

    def __update_layout_property_label(self):
        self.__update_label(self.layout_prop)

    def __update_property_label(self):
        self.__update_label(self.prop)

        if not self.bindable:
            return

        if self.prop.bind_property_id or self.prop.binding_expression_id:
            self.bind_icon.props.icon_name = "cmb-binded-symbolic"
            self.remove_css_class("hidden")
        else:
            self.bind_icon.props.icon_name = "cmb-bind-symbolic"
            self.add_css_class("hidden")

    def __on_bind_button_clicked(self, button):
        popover = CmbBindingPopover(prop=self.prop)
        popover.set_parent(self)

        # Destroy popup on close
        popover.connect("closed", lambda p: p.unparent())
        popover.popup()

    def __on_click_gesture_released(self, gesture, n_press, x, y):
        if self.prop is None:
            return

        if n_press == 1 and gesture.get_current_button() == 3 and \
           x >= 0 and x <= self.get_width() and y >= 0 and y <= self.get_height():
            self.menu.popup()

    @Gtk.Template.Callback("on_reset_property_clicked")
    def __on_reset_property_clicked(self, button):
        self.prop.reset()
        self.menu.popdown()

    @Gtk.Template.Callback("on_serialize_default_toggled")
    def __on_serialize_default_property_toggled(self, check):
        if self.__on_init:
            return

        self.prop.serialize_default_value = check.props.active
        self.menu.popdown()


Gtk.WidgetClass.set_css_name(CmbPropertyLabel, "CmbPropertyLabel")



# Merengue Application
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
import gi
import logging

gi.require_version('GIRepository', '3.0')

from . import config
from gi.repository import GLib, Gio, Gtk


resource = Gio.Resource.load(os.path.join(config.pkgdatadir, "merengue.gresource"))
resource._register()

repository = gi.Repository.get_default()
repository.prepend_search_path(config.privatecambalachedir)
repository.prepend_library_path(config.privatecambalachedir)

gi.require_version("CambalachePrivate", "4.0" if Gtk.MAJOR_VERSION == 4 else "3.0")
from gi.repository import CambalachePrivate

def __log_writer_handler(level, field_list, data):
    fields = {f.key: CambalachePrivate.log_field_get_string(f) for f in field_list if f.length < 0}

    if fields.get("GLIB_DOMAIN") == "GLib-GIO":
        if fields.get("MESSAGE").startswith("Adding GResources overlay") or \
           fields.get("MESSAGE").startswith("Mapped file") or \
           fields.get("MESSAGE").startswith("Can't mmap overlay file"):
            return GLib.LogWriterOutput.HANDLED

    return GLib.log_writer_default(level, field_list, data)


GLib.log_set_writer_func(__log_writer_handler)


def getLogger(name):
    formatter = logging.Formatter("%(levelname)s:%(name)s %(message)s")

    ch = logging.StreamHandler()
    ch.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(os.environ.get("MERENGUE_LOGLEVEL", "WARNING").upper())
    logger.addHandler(ch)

    return logger


from .mrg_application import MrgApplication
from .mrg_controller import MrgController
from .mrg_placeholder import MrgPlaceholder
from . import utils




#
# CmbNewProjectDialog
#
# Copyright (C) 2026  Matthieu Lorier
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Matthieu Lorier <loriermatthieu@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os

from gi.repository import Gtk, Adw, GObject, GLib
from cambalache import getLogger, _, utils

logger = getLogger(__name__)


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/app/cmb_np_dialog.ui")
class CmbNewProjectDialog(Adw.Dialog):
    __gtype_name__ = "CmbNewProjectDialog"
    __gsignals__ = {
        "create-new-project": (GObject.SignalFlags.ACTION, None, (str, str, str))
    }

    name_entry = Gtk.Template.Child()
    ui_filename_entry = Gtk.Template.Child()
    location_row = Gtk.Template.Child()
    toolkit_chooser = Gtk.Template.Child()
    create_button = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        home = GLib.get_home_dir()
        projects = os.path.join(home, "Projects")
        self.location = projects if os.path.isdir(projects) else home

    @GObject.Property(type=str)
    def location(self):
        return self.__location

    @location.setter
    def location(self, location):
        self.__location = location
        self.location_row.props.subtitle = utils.friendly_homedir(location)

    @Gtk.Template.Callback("on_name_entry_changed")
    def __on_name_entry_changed(self, entry):
        name = entry.get_text()
        sensitive = len(name) > 0
        filename = ""

        if sensitive:
            name, ext = os.path.splitext(name)
            filename = os.path.join(self.location, name + ".cmb")

            if os.path.exists(filename):
                error_msg = _("File name already exists, choose a different name.")
                self.name_entry.props.tooltip_text = error_msg
                self.create_button.props.tooltip_text = error_msg

                self.name_entry.add_css_class("warning")

                self.create_button.set_sensitive(False)
            else:
                self.name_entry.remove_css_class("warning")
                self.name_entry.props.tooltip_text = None
                self.create_button.props.tooltip_text = None
                self.create_button.set_sensitive(True)

            self.ui_filename_entry.set_text(name + ".ui")
        else:
            self.name_entry.add_css_class("warning")
            self.ui_filename_entry.set_text("")

        self.location_row.set_sensitive(sensitive)
        self.ui_filename_entry.set_sensitive(sensitive)

    @Gtk.Template.Callback("on_location_button_clicked")
    def __on_location_button_clicked(self, button):
        def dialog_callback(dialog, res):
            try:
                self.location = dialog.select_folder_finish(res).get_path()
            except Exception as e:
                logger.warning(f"Error {e}")

        dialog = Gtk.FileDialog(title=_("Select project location"), modal=True)
        dialog.select_folder(self.props.root, None, dialog_callback)

    @Gtk.Template.Callback("on_create_button_clicked")
    def __on_create_button_clicked(self, button):
        name = self.name_entry.props.text
        uiname = self.ui_filename_entry.props.text
        filename = ""
        uipath = ""
        target_tk = self.toolkit_chooser.props.active_name

        if len(name):
            name, ext = os.path.splitext(name)
            filename = os.path.join(self.location, name + ".cmb")

            if len(uiname) == 0:
                uiname = self.name_entry.props.text + ".ui"

            if os.path.exists(filename):
                self.name_entry.props.title = _("File name already exists, choose a different name.")
                self.set_focus(self.name_entry)
                return

            uipath = os.path.join(self.location, uiname)

        self.emit("create-new-project", target_tk, filename, uipath)



#!/usr/bin/pytest

"""
import .ui files into cambalache and compare it to blueprint compiler output
"""
import os
import pytest

from lxml import etree
from cambalache import CmbProject
from cambalache.cmb_blueprint import cmb_blueprint_decompile, cmb_blueprint_compile, CmbBlueprintUnsupportedError


def tostring(ui):
    db = ui.project.db

    # Internal API to ensure Cambalache outputs the same as blueprint compiler
    db._output_lowercase_boolean = True
    db._output_use_enum_value = True

    tree = db.export_ui(ui.ui_id)

    if tree is None:
        return None

    root = tree.getroot()

    # Remove all XML comments since they are not supported by blueprint
    for node in root.xpath("//comment()"):
        parent = node.getparent()
        if parent is not None:
            parent.remove(node)

    for node in root.iterfind("requires"):
        lib = node.get("lib", None)
        if lib != "gtk":
            root.remove(node)

    return etree.tostring(root, pretty_print=True, encoding="UTF-8").decode("UTF-8")


def get_exported_and_blueprint(filename):
    """
    import .ui file and compare it with the exported version
    """
    path = os.path.join(os.path.dirname(__file__), "gtk-4.0", filename)
    project = CmbProject(target_tk="gtk-4.0")
    ui, msgs, detail_msg = project.import_file(path)

    assert (ui)
    assert (msgs is None)
    assert (detail_msg is None)

    # Export Cambalache UI
    str_exported = tostring(ui)

    # Decompile and recompile UI to Blueprint
    blueprint_decompiled = cmb_blueprint_decompile(str_exported)
    blueprint_compiled = cmb_blueprint_compile(blueprint_decompiled)
    assert blueprint_compiled is not None

    # Remove blueprint DO NOT EDIT comment
    root = etree.fromstring(blueprint_compiled)
    blueprint_compiled = etree.tostring(root, pretty_print=True, encoding="UTF-8").decode("UTF-8")

    return str_exported, blueprint_compiled


@pytest.mark.parametrize("filename", [
    "window.ui",
    "children.ui",
    "layout.ui",
    "signals.ui",
    "template.ui",
    "inline_object.ui",
    "stack_page.ui",
    "comboboxtext.ui",
    "style.ui",
    "filefilter.ui",
    "menu.ui",

    # help-text is not an accessibility property
    "accessibility.ui",

    # bind-flags missing
    # "bindings.ui",

    # Translation comment missing
    # "string_list.ui",
])
def test_assert_exported_and_compiled(filename):
    """
    import .ui file and compare it with the exported version
    """
    str_exported, blueprint_compiled = get_exported_and_blueprint(filename)

    # Compare blueprint generated string with cambalache
    assert blueprint_compiled.strip() == str_exported.strip()


@pytest.mark.parametrize("filename", [
    "liststore.ui",
    "treestore.ui",
    "label.ui",
])
def test_assert_exported_and_compiled_unsupported(filename):
    with pytest.raises(CmbBlueprintUnsupportedError):
        str_exported, blueprint_compiled = get_exported_and_blueprint(filename)




#
# CmbCatalogRow - Cambalache Catalog Row
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk, Adw
from .cmb_library_info import CmbLibraryInfo
from cambalache import ngettext


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_library_row.ui")
class CmbLibraryRow(Adw.PreferencesRow):
    __gtype_name__ = "CmbLibraryRow"

    info = GObject.Property(type=CmbLibraryInfo, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    label = Gtk.Template.Child()
    description = Gtk.Template.Child()
    stack = Gtk.Template.Child()
    clear = Gtk.Template.Child()
    switch = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self.n_objects = 0

        super().__init__(**kwargs)

        self.switch.props.active = self.info.enabled
        self.__init_n_objects()

    def __init_n_objects(self):
        info = self.info

        self.n_objects = info.project.db.execute(
            "SELECT COUNT(object_id) FROM object, type WHERE object.type_id = type.type_id AND type.library_id=?;",
            (info.library_id, )
        ).fetchone()[0]

        if self.n_objects > 0:
            self.label.props.label = ngettext(
                "<b>{catalog}-{version}</b> <small>(used by {n} object)</small>",
                "<b>{catalog}-{version}</b> <small>(used by {n} objects)</small>",
                self.n_objects
            ).format(
                catalog=info.library_id,
                version=info.version,
                n=self.n_objects
            )
            self.stack.set_visible_child(self.clear)
        else:
            self.label.props.label = f"{info.library_id}-{info.version}"
            self.stack.set_visible_child(self.switch)

    @Gtk.Template.Callback("on_clear_clicked")
    def __on_clear_clicked(self, button):
        dialog = Adw.AlertDialog(
            heading=ngettext(
                "Delete {n} object?",
                "Delete {n} objects?",
                self.n_objects
            ).format(n=self.n_objects),
            body=ngettext(
                "This project contains {n} object from {catalog}",
                "This project contains {n} objects from {catalog}",
                self.n_objects
            ).format(
                catalog=self.info.library_id,
                n=self.n_objects
            ),
        )
        dialog.add_response("cancel", "Cancel")
        dialog.add_response("delete", "Delete")
        dialog.set_response_appearance("delete", Adw.ResponseAppearance.DESTRUCTIVE)

        dialog.set_default_response("cancel")
        dialog.set_close_response("cancel")

        def on_dialog_response(obj, response):
            if response == "delete":
                self.info.project.remove_library_objects(self.info)
                self.__init_n_objects()

        dialog.connect("response", on_dialog_response)
        dialog.present(self.props.root)

    @Gtk.Template.Callback("on_switch_notify")
    def __on_switch_notify(self, switch, pspec):
        # Load/Unload catalog
        self.info.enabled = switch.props.active

        # Sync status in case unload failed
        if switch.props.active != self.info.enabled:
            switch.props.active = self.info.enabled


Gtk.WidgetClass.set_css_name(CmbLibraryRow, "CmbLibraryRow")



#
# CmbFileButton
#
# Copyright (C) 2021-2023  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os

from cambalache import _
from gi.repository import GObject, Gio, Gtk


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/control/cmb_file_button.ui")
class CmbFileButton(Gtk.Button):
    __gtype_name__ = "CmbFileButton"

    dirname = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    dialog_title = GObject.Property(type=str, default=_("Select filename"), flags=GObject.ParamFlags.READWRITE)
    accept_label = GObject.Property(type=str, default=_("Select"), flags=GObject.ParamFlags.READWRITE)
    unnamed_filename = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    use_open = GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READWRITE)

    label = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__filename = None
        self.__filters = None

    @Gtk.Template.Callback("on_button_clicked")
    def __on_button_clicked(self, button):
        dialog = Gtk.FileDialog(
            modal=True,
            filters=self.__filters,
            title=self.dialog_title,
            accept_label=self.accept_label
        )

        if self.dirname is not None:
            if self.__filename:
                fullpath = os.path.join(self.dirname, self.__filename)

                file = Gio.File.new_for_path(fullpath)
                dialog.set_initial_file(file)

                # See which filter matches the file info and use it as default
                if self.__filters and file.query_exists(None):
                    info = file.query_info(Gio.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE, Gio.FileQueryInfoFlags.NONE, None)
                    for filter in self.__filters:
                        if filter.match(info):
                            dialog.set_default_filter(filter)
                            break
            else:
                dialog.set_initial_folder(Gio.File.new_for_path(self.dirname))
                if self.unnamed_filename:
                    dialog.set_initial_name(self.unnamed_filename)

        def dialog_callback(dialog, res):
            try:
                file = dialog.open_finish(res) if self.use_open else dialog.save_finish(res)
                self.cmb_value = os.path.relpath(file.get_path(), start=self.dirname)
            except Exception:
                pass

        if self.use_open:
            dialog.open(self.get_root(), None, dialog_callback)
        else:
            dialog.save(self.get_root(), None, dialog_callback)

    @GObject.Property(type=str)
    def cmb_value(self):
        return self.__filename

    @cmb_value.setter
    def _set_cmb_value(self, value):
        if value == self.__filename:
            return

        self.__filename = value if value is not None else ""
        self.label.set_label(self.__filename)

    @GObject.Property(type=str)
    def mime_types(self):
        if self.__filters:
            return ";".join([f.props.mime_types for f in self.__filters])
        return ""

    @mime_types.setter
    def _set_mime_types(self, value):
        if value:
            self.__filters = Gio.ListStore()
            for mime in value.split(';'):
                self.__filters.append(Gtk.FileFilter(mime_types=[mime]))
        else:
            self.__filters = None



#
# CmbColumnView
#
# Copyright (C) 2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GLib, GObject, Gdk, Gtk

from .cmb_ui import CmbUI
from .cmb_object import CmbObject
from .cmb_gresource import CmbGResource
from .cmb_context_menu import CmbContextMenu
from .cmb_path import CmbPath
from .cmb_project import CmbProject
from .cmb_tree_expander import CmbTreeExpander
from .cmb_base_file_monitor import CmbBaseFileMonitor, FileStatus

from cambalache import _


class CmbListView(Gtk.ListView):
    __gtype_name__ = "CmbListView"

    def __init__(self, **kwargs):
        self.__project = None
        self.__tree_model = None
        self.__in_selection_change = False
        self.single_selection = Gtk.SingleSelection()

        super().__init__(**kwargs)

        self.props.has_tooltip = True
        self.props.hexpand = True

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_factory_setup)
        factory.connect("bind", self._on_factory_bind)
        factory.connect("unbind", self._on_factory_unbind)
        self.props.factory = factory

        self.single_selection.connect("notify::selected-item", self.__on_selected_item_notify)
        self.set_model(self.single_selection)

        gesture = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        gesture.connect("pressed", self.__on_button_press)
        self.add_controller(gesture)

        self.connect("activate", self.__on_activate)

        self.add_css_class("navigation-sidebar")
        self.add_css_class("cmb-list-view")

    def __on_button_press(self, gesture, npress, x, y):
        expander = self.__get_tree_expander(x, y)

        if expander is None or npress != 1:
            return False

        # Select row at x,y
        list_row = expander.get_list_row()
        self.single_selection.set_selected(list_row.get_position())

        menu = CmbContextMenu()

        if self.__project:
            menu.target_tk = self.__project.target_tk

        menu.set_parent(self)
        menu.popup_at(x, y)
        return True

    def __get_path_parent(self, obj):
        if isinstance(obj, CmbObject):
            parent = obj.parent
            return parent if parent else obj.ui
        elif isinstance(obj, CmbGResource):
            return obj.path_parent if obj.resource_type == "gresources" else obj.parent
        elif isinstance(obj, CmbProject):
            return None

        return obj.path_parent

    def __get_object_ancestors(self, obj):
        ancestors = set()

        parent = self.__get_path_parent(obj)
        while parent:
            ancestors.add(parent)
            parent = self.__get_path_parent(parent)

        return ancestors

    def __object_ancestor_expand(self, obj):
        ancestors = self.__get_object_ancestors(obj)
        i = 0

        # Iterate over tree model
        # NOTE: only visible/expanded rows are returned
        list_row = self.__tree_model.get_row(i)
        while list_row:
            item = list_row.get_item()

            # Return position if we reached the object row
            if item == obj:
                return i
            elif item in ancestors:
                # Expand row if its part of the hierarchy
                list_row.set_expanded(True)

            i += 1
            list_row = self.__tree_model.get_row(i)

        return None

    def __on_project_selection_changed(self, p):
        list_row = self.single_selection.get_selected_item()
        current_selection = [list_row.get_item()] if list_row else []
        selection = self.__project.get_selection()

        if selection == current_selection:
            return

        self.__in_selection_change = True

        if len(selection) > 0:
            position = self.__object_ancestor_expand(selection[0])
            if position is not None:
                self.single_selection.select_item(position, True)
            else:
                self.single_selection.unselect_all()
        else:
            self.single_selection.unselect_all()

        self.__in_selection_change = False

    @GObject.Property(type=CmbProject)
    def project(self):
        return self.__project

    @project.setter
    def _set_project(self, project):
        if self.__project:
            self.__project.disconnect_by_func(self.__on_project_selection_changed)

        self.__project = project

        if project:
            self.__tree_model = Gtk.TreeListModel.new(
                project,
                False,
                False,
                self.__tree_model_create_func,
                None
            )
            self.single_selection.props.model = self.__tree_model
            self.__project.connect("selection-changed", self.__on_project_selection_changed)
        else:
            self.single_selection.props.model = None

    def __tree_model_create_func(self, item, data):
        if isinstance(item, CmbObject):
            return item
        elif isinstance(item, CmbUI):
            return item
        elif isinstance(item, CmbGResource):
            return item
        elif isinstance(item, CmbPath):
            return item

        return None

    def __on_selected_item_notify(self, single_selection, pspec):
        if self.__in_selection_change or self.__project is None:
            return

        list_item = single_selection.get_selected_item()

        if list_item is None:
            self.__project.set_selection([])
            return

        if not list_item.get_expanded():
            list_item.set_expanded(True)

        item = list_item.get_item()

        if item and not isinstance(item, CmbPath):
            item = list_item.get_item()
            self.__project.set_selection([item])
        else:
            self.__project.set_selection([])

    def _on_factory_setup(self, factory, list_item):
        expander = CmbTreeExpander()
        list_item.set_child(expander)

    def _on_factory_bind(self, factory, list_item):
        row = list_item.get_item()
        expander = list_item.get_child()
        expander.set_list_row(row)
        expander.update_bind()

    def _on_factory_unbind(self, factory, list_item):
        expander = list_item.get_child()
        expander.clear_bind()

    def __get_tree_expander(self, x, y):
        pick = self.pick(x, y, Gtk.PickFlags.DEFAULT)

        if pick is None:
            return None

        if isinstance(pick, Gtk.TreeExpander):
            return pick

        child = pick.get_first_child()

        if child and isinstance(child, Gtk.TreeExpander):
            return child

        parent = pick.props.parent
        if parent and isinstance(parent, Gtk.TreeExpander):
            return parent

        return None

    def __on_activate(self, column_view, position):
        item = self.__tree_model.get_item(position)
        item.set_expanded(not item.get_expanded())

    def do_query_tooltip(self, x, y, keyboard_mode, tooltip):
        expander = self.__get_tree_expander(x, y)

        if expander is None:
            return False

        obj = expander.get_item()

        msg = None
        if isinstance(obj, CmbObject):
            msg = obj.version_warning
        elif isinstance(obj, CmbBaseFileMonitor):
            status = obj.file_status

            if status == FileStatus.CHANGED:
                msg = _("File Has Changed on Disk")
            elif status == FileStatus.NOT_FOUND:
                msg = _("File Not Found")
            elif status == FileStatus.DELETED:
                msg = _("File Deleted")
            elif status == FileStatus.RENAMED:
                msg = _("File renamed to {new_filename}").format(new_filename=obj.new_filename)

        if msg:
            tooltip.set_text(msg)
            return True

        return False



#
# CmbFlagsEntry
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk
from ..cmb_type_info import CmbTypeInfo


class CmbFlagsEntry(Gtk.Entry):
    __gtype_name__ = "CmbFlagsEntry"

    info = GObject.Property(type=CmbTypeInfo, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)
    id_column = GObject.Property(type=int, default=1, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)
    text_column = GObject.Property(type=int, default=1, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)
    value_column = GObject.Property(type=int, default=2, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)
    separator = GObject.Property(type=str, default="|", flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT)

    def __init__(self, **kwargs):
        self.flags = {}
        self._checks = {}
        self._popover = None

        super().__init__(**kwargs)

        self.props.editable = False
        self.props.secondary_icon_name = "document-edit-symbolic"

        self.connect("icon-release", self.__on_icon_release)

    def __init_popover(self):
        if self._popover:
            return

        self._popover = Gtk.Popover()
        self._popover.set_parent(self)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.append(Gtk.Label(label=f"<b>{self.info.type_id}</b>", use_markup=True))
        box.append(Gtk.Separator())
        sw = Gtk.ScrolledWindow(hscrollbar_policy=Gtk.PolicyType.NEVER, propagate_natural_height=True, max_content_height=360)
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        sw.set_child(vbox)
        box.append(sw)

        for row in self.info.flags:
            flag = row[self.text_column]
            flag_id = row[self.id_column]

            check = Gtk.CheckButton(label=flag)
            check.connect("toggled", self.__on_check_toggled, flag_id)
            vbox.append(check)
            self._checks[flag_id] = check

        self._popover.set_child(box)

    def __on_check_toggled(self, check, flag_id):
        self.flags[flag_id] = check.props.active
        self.props.text = self.__to_string()
        self.notify("cmb-value")

    def __on_icon_release(self, obj, pos):
        self.__init_popover()
        self._popover.popup()

    def __to_string(self):
        retval = None
        for row in self.info.flags:
            flag_id = row[self.id_column]
            if self.flags.get(flag_id, False):
                retval = flag_id if retval is None else f"{retval} {self.separator} {flag_id}"

        return retval if retval is not None else ""

    @GObject.Property(type=str)
    def cmb_value(self):
        return self.props.text if self.props.text != "" else None

    @cmb_value.setter
    def _set_cmb_value(self, value):
        if value == self.props.text:
            return

        self.props.text = value if value is not None else ""

        self.flags = {}
        for check in self._checks:
            self._checks[check].props.active = False

        if value:
            tokens = [t.strip() for t in value.split(self.separator)]

            for row in self.info.flags:
                flag_id = row[self.id_column]
                flag_name = row[0]
                flag_nick = row[1]

                check = self._checks.get(flag_id, None)
                if check:
                    val = flag_name in tokens or flag_nick in tokens
                    check.props.active = val
                    self.flags[flag_id] = val



#
# CmbIconNameEntry
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os

from cambalache import _
from gi.repository import GLib, GdkPixbuf, GObject, Gdk, Gtk, Pango
from .cmb_entry import CmbEntry
from .icon_naming_spec import standard_icon_context, standard_icon_names


class CmbIconNameEntry(CmbEntry):
    __gtype_name__ = "CmbIconNameEntry"

    object = GObject.Property(type=GObject.Object, flags=GObject.ParamFlags.READWRITE)

    standard_only = GObject.Property(type=bool, default=True, flags=GObject.ParamFlags.READWRITE)
    symbolic_only = GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READWRITE)

    COL_ICON_NAME = 0
    COL_MARKUP = 1
    COL_CONTEXT = 2
    COL_STANDARD = 3
    COL_SYMBOLIC = 4
    COL_STANDARD_SYMBOLIC = 5
    COL_PIXBUF = 6

    # Model, store it in a Python class variable to share between all instances
    icon_model = None

    iconlist = []

    def __init__(self, **kwargs):
        self._filters = {}

        super().__init__(**kwargs)

        GObject.Object.bind_property(self, "cmb-value", self, "primary-icon-name", GObject.BindingFlags.SYNC_CREATE)

        self.props.secondary_icon_name = "document-edit-symbolic"
        self.connect("icon-press", self.__on_icon_pressed)

        CmbIconNameEntry.ensure_icon_model()

        # Completion
        self.__completion = Gtk.EntryCompletion()
        self.__completion.props.model = self.icon_model
        self.__completion.props.text_column = self.COL_ICON_NAME
        self.__completion.props.inline_completion = True
        self.__completion.props.inline_selection = True
        self.__completion.props.popup_set_width = True
        self.props.completion = self.__completion

        self.__completion.set_match_func(lambda o, key, iter, d: key in self.icon_model[iter][0], None)

        # Icon
        renderer_text = Gtk.CellRendererPixbuf(xpad=4)
        self.__completion.pack_start(renderer_text, False)
        self.__completion.add_attribute(renderer_text, "icon-name", 0)

        # Icon Name
        renderer_text = Gtk.CellRendererText(ellipsize=Pango.EllipsizeMode.END)
        self.__completion.pack_start(renderer_text, False)
        self.__completion.add_attribute(renderer_text, "markup", 1)

    @classmethod
    def ensure_icon_model(cls):
        if cls.icon_model:
            return

        # icon-name markup context standard symbolic standard_symbolic pixbuf
        cls.icon_model = Gtk.ListStore(str, str, str, bool, bool, bool, GdkPixbuf.Pixbuf)

        iconlist = []

        theme = Gtk.IconTheme.get_for_display(Gdk.Display.get_default())

        # FIXME: get the context/category of each icon
        for icon in theme.get_icon_names():
            iconlist.append((icon, "cmb_all", icon in standard_icon_names))

        for icon, context, standard in sorted(iconlist, key=lambda i: i[0].lower()):
            if icon.endswith(".symbolic"):
                continue

            icon_paintable = theme.lookup_icon(icon, None, 32, 1, Gtk.TextDirection.NONE, Gtk.IconLookupFlags.PRELOAD)
            symbolic = icon_paintable.is_symbolic()

            icon_file = icon_paintable.get_file()
            if icon_file is None:
                continue

            icon_path = icon_file.get_path()
            if icon_path is None or not os.path.exists(icon_path):
                continue

            standard_symbolic = symbolic and icon.removesuffix("-symbolic") in standard_icon_names

            try:
                iter = cls.icon_model.append(
                    [icon, icon if standard else f"<i>{icon}</i>", context, standard, symbolic, standard_symbolic, None]
                )
                cls.iconlist.append((icon_file, iter))
            except Exception as e:
                print(e)

        # Kickoff async loading
        file, iter = cls.iconlist.pop()
        file.read_async(GLib.PRIORITY_DEFAULT, None, cls.__load_file_finish, iter)

    @classmethod
    def __load_file_finish(cls, obj, res, iter):
        stream = obj.read_finish(res)
        GdkPixbuf.Pixbuf.new_from_stream_at_scale_async(stream, 32, 32, True, None, cls.__load_icon_finish, iter)

    @classmethod
    def __load_icon_finish(cls, obj, res, iter):
        try:
            cls.icon_model[iter][cls.COL_PIXBUF] = GdkPixbuf.Pixbuf.new_from_stream_finish(res)
        except Exception as e:
            print(e)

        if len(cls.iconlist):
            file, iter = cls.iconlist.pop()
            file.read_async(GLib.PRIORITY_DEFAULT, None, cls.__load_file_finish, iter)

    def __model_filter_func(self, model, iter, data):
        if self.standard_only and self.symbolic_only:
            if not model[iter][self.COL_STANDARD_SYMBOLIC]:
                return False
        elif self.standard_only and not model[iter][self.COL_STANDARD]:
            return False
        elif self.symbolic_only and not model[iter][self.COL_SYMBOLIC]:
            return False

        if data == "cmb_all":
            return True

        return model[iter][self.COL_CONTEXT] == data

    def __on_check_active_notify(self, button, pspec):
        for filter in self._filters:
            self._filters[filter].refilter()

    def __on_view_selection_changed(self, view):
        selection = view.get_selected_items()

        if selection:
            model = view.props.model
            iter = model.get_iter(selection[0])
            self.cmb_value = model[iter][self.COL_ICON_NAME]
        else:
            self.cmb_value = None

    def __on_icon_pressed(self, widget, icon_pos):
        # Create popover with icon chooser
        popover = Gtk.Popover()
        popover.set_parent(self)

        hbox = Gtk.Box(visible=True)
        vbox = Gtk.Box(visible=True, orientation=Gtk.Orientation.VERTICAL, vexpand=True)
        stack = Gtk.Stack(visible=True, transition_type=Gtk.StackTransitionType.CROSSFADE)
        sidebar = Gtk.StackSidebar(visible=True, stack=stack, vexpand=True)
        vbox.append(sidebar)
        hbox.append(vbox)
        hbox.append(stack)

        # sorted_contexts = sorted(theme.list_contexts())
        sorted_contexts = []
        sorted_contexts.insert(0, "cmb_all")

        # Add one icon view per context
        for context in sorted_contexts:
            filter = self._filters.get(context, None)

            if filter is None:
                self._filters[context] = Gtk.TreeModelFilter(child_model=self.icon_model)
                filter = self._filters[context]
                filter.set_visible_func(self.__model_filter_func, data=context)
                filter.refilter()

            sw = Gtk.ScrolledWindow(visible=True, min_content_width=600, min_content_height=480)
            view = Gtk.IconView(visible=True, model=filter, pixbuf_column=self.COL_PIXBUF, text_column=self.COL_ICON_NAME)
            view.connect("selection-changed", self.__on_view_selection_changed)
            sw.set_child(view)
            stack.add_titled(sw, context, standard_icon_context.get(context, context))

        # Add filters
        for prop, label in [("standard_only", _("Only standard")), ("symbolic_only", _("Only symbolic"))]:
            check = Gtk.CheckButton(visible=True, label=label)
            GObject.Object.bind_property(
                self, prop, check, "active", GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL
            )
            check.connect_after("notify::active", self.__on_check_active_notify)
            vbox.append(check)

        popover.add_css_class("cmb-icon-chooser")
        popover.set_child(hbox)
        popover.popup()



# GtkNotebook Controller
#
# Copyright (C) 2022  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .mrg_gtk_widget import MrgGtkWidget
from merengue import MrgPlaceholder

from merengue import getLogger

logger = getLogger(__name__)


class MrgGtkNotebook(MrgGtkWidget):
    object = GObject.Property(type=Gtk.Notebook, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.current_page = 0

        super().__init__(**kwargs)

    def get_children(self):
        if Gtk.MAJOR_VERSION == 4:
            return list(self.object.get_pages() if self.object else [])

        return super().get_children()

    def __get_placeholder(self):
        for child in self.get_children():
            if Gtk.MAJOR_VERSION == 4 and isinstance(child, Gtk.NotebookPage):
                child = child.get_child()

            if isinstance(child, MrgPlaceholder):
                return child
        return None

    def __ensure_placeholders(self):
        if self.object is None:
            return

        if len(self.get_children()) == 0:
            self.add(MrgPlaceholder(visible=True, controller=self))

    def __on_object_page_notify(self, obj, pspec):
        if obj.is_visible():
            self.current_page = obj.props.page

    def __on_object_map(self, obj):
        self.object.props.page = self.current_page

    def object_changed(self, old, new):
        super().object_changed(old, new)

        if old:
            try:
                old.disconnect_by_func(self.__on_object_page_notify)
                old.disconnect_by_func(self.__on_object_map)
            except Exception:
                pass

        self.__ensure_placeholders()

        if self.object:
            self.object.connect("notify::page", self.__on_object_page_notify)
            self.object.connect("map", self.__on_object_map)

    def show_child(self, child):
        if Gtk.MAJOR_VERSION == 4:
            if isinstance(child, Gtk.NotebookPage):
                position = child.props.position
            else:
                position = self.object.page_num(child)
        else:
            position = self.object.page_num(child)

        if position >= 0:
            self.object.set_current_page(position)

    def add(self, child):
        if self.object is None:
            return

        if Gtk.MAJOR_VERSION == 3:
            self.object.add(child)
        else:
            self.object.append_page(child, None)

    def remove_child(self, child):
        if self.object is None:
            return

        if Gtk.MAJOR_VERSION == 4:
            self.object.detach_tab(child)
        else:
            super().remove_child(child)

    def add_placeholder(self, mod):
        placeholder = self.__get_placeholder()

        if placeholder is None:
            placeholder = MrgPlaceholder(visible=True, controller=self)
            self.add(placeholder)

        self.show_child(placeholder)

    def remove_placeholder(self, mod):
        placeholder = self.__get_placeholder()
        if placeholder:
            self.remove_child(placeholder)



#
# CmbChildTypeComboBox
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk


class CmbChildTypeComboBox(Gtk.ComboBox):
    __gtype_name__ = "CmbChildTypeComboBox"

    object = GObject.Property(type=GObject.Object, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.connect("changed", self.__on_changed)

        # Model, store it in a Python variable to make sure we hold a reference
        # First column is the ID and the second is if you can select the child or not
        self.__model = Gtk.ListStore(str, bool)
        self.props.model = self.__model
        self.props.id_column = 0

        # Simple cell renderer
        renderer_text = Gtk.CellRendererText()
        self.pack_start(renderer_text, True)
        self.add_attribute(renderer_text, "text", 0)
        self.add_attribute(renderer_text, "sensitive", 1)

        self.__populate_model()

    def __populate_model(self):
        self.__model.clear()

        parent = self.object.parent
        if parent is None:
            return

        self.__model.append([None, True])

        pinfo = parent.info
        while pinfo:
            if pinfo.child_types:
                for t in pinfo.child_types:
                    self.__model.append([t, True])
            pinfo = pinfo.parent

    def __on_changed(self, obj):
        self.notify("cmb-value")

    @GObject.Property(type=str)
    def cmb_value(self):
        return self.props.active_id

    @cmb_value.setter
    def _set_cmb_value(self, value):
        self.props.active_id = value



# HdySearchBar Controller
#
# Copyright (C) 2022  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk, Handy

from merengue import utils
from merengue.mrg_gtk import MrgGtkBin


class MrgHdySearchBar(MrgGtkBin):
    object = GObject.Property(type=Handy.SearchBar, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_children(self):
        if not self.object:
            return []

        revealer = self.object.get_child()
        if not issubclass(type(revealer), Gtk.Revealer):
            return []

        box = revealer.get_child()
        if not issubclass(type(box), Gtk.Box):
            return []

        for child in box.get_children():
            name = utils.object_get_builder_id(child)

            if name in ["start", "end"]:
                continue

            return [child]

        return []

    def show_child(self, child):
        if self.object:
            self.object.props.search_mode_enabled = True



#
# Cambalache controller registry
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import inspect
from gi.repository import GObject

from .mrg_controller import MrgController


class MrgControllerRegistry(GObject.GObject):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Loaded modules
        self.modules = {}

        # Registered controllers
        self.registry = {}

    # Object set property wrapper
    def load_module(self, namespace, module):
        def get_object_type(klass):
            props = klass.list_properties()

            for pspec in props:
                if pspec.name == "object":
                    return pspec.value_type

            return None

        self.modules[namespace] = module

        for component in module.__dict__.items():
            name, klass = component
            if inspect.isclass(klass) and issubclass(klass, MrgController):
                object_type = get_object_type(klass)
                self.registry[object_type] = klass

    def new_controller_for_type(self, gtype, app):
        klass = None

        while gtype and klass is None:
            klass = self.registry.get(gtype, None)
            try:
                gtype = GObject.type_parent(gtype)
            except Exception:
                gtype = None
                break

        return klass(app=app) if klass else MrgController(app=app)



#
# CmbSignalEditor - Cambalache Signal Editor
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk, Pango

from .cmb_object import CmbObject
from . import utils

from enum import Enum


class Col(Enum):
    SIGNAL = 0
    OWNER_ID = 1
    SIGNAL_ID = 2
    DETAIL = 3
    HANDLER = 4
    USER_DATA = 5
    SWAP = 6
    AFTER = 7
    INFO = 8
    WARNING_MESSAGE = 9


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_signal_editor.ui")
class CmbSignalEditor(Gtk.Box):
    __gtype_name__ = "CmbSignalEditor"

    treeview = Gtk.Template.Child()
    treestore = Gtk.Template.Child()

    signal_id_column = Gtk.Template.Child()
    signal_id = Gtk.Template.Child()
    handler_column = Gtk.Template.Child()
    handler = Gtk.Template.Child()
    user_data_column = Gtk.Template.Child()
    user_data = Gtk.Template.Child()
    swap_column = Gtk.Template.Child()
    swap = Gtk.Template.Child()
    after_column = Gtk.Template.Child()
    after = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self._object = None
        super().__init__(**kwargs)

        self.signal_id_column.set_cell_data_func(self.signal_id, self.__signal_id_data_func, None)
        self.handler_column.set_cell_data_func(self.handler, self.__data_func, Col.HANDLER.value)
        self.user_data_column.set_cell_data_func(self.user_data, self.__data_func, Col.USER_DATA.value)
        self.swap_column.set_cell_data_func(self.swap, self.__data_func, Col.SWAP.value)
        self.after_column.set_cell_data_func(self.after, self.__data_func, Col.AFTER.value)

    @GObject.Property(type=CmbObject)
    def object(self):
        return self._object

    @object.setter
    def _set_object(self, obj):
        if self._object:
            self.treestore.clear()
            self._object.disconnect_by_func(self.__on_signal_added)
            self._object.disconnect_by_func(self.__on_signal_removed)
            self._object.disconnect_by_func(self.__on_signal_changed)

        self._object = obj

        if obj:
            self.__populate_treestore()
            self._object.connect("signal-added", self.__on_signal_added)
            self._object.connect("signal-removed", self.__on_signal_removed)
            self._object.connect("signal-changed", self.__on_signal_changed)

    @Gtk.Template.Callback("on_handler_edited")
    def __on_handler_edited(self, renderer, path, new_text):
        iter_ = self.treestore.get_iter(path)
        signal = self.treestore[iter_][Col.SIGNAL.value]

        if signal is None:
            if len(new_text) > 0:
                owner_id = self.treestore[iter_][Col.OWNER_ID.value]
                signal_id = self.treestore[iter_][Col.SIGNAL_ID.value]
                self._object.add_signal(owner_id, signal_id, new_text)
        else:
            self.treestore[iter_][Col.HANDLER.value] = new_text
            if len(new_text) > 0:
                signal.handler = new_text
            else:
                self._object.remove_signal(signal)

    @Gtk.Template.Callback("on_detail_edited")
    def __on_detail_edited(self, renderer, path, new_text):
        iter_ = self.treestore.get_iter(path)
        signal = self.treestore[iter_][Col.SIGNAL.value]

        if signal is not None:
            if len(new_text) > 0:
                tokens = new_text.split("::")
                if len(tokens) == 2 and len(tokens[1]) > 0:
                    signal.detail = tokens[1]
                else:
                    signal.detail = None
            else:
                signal.detail = None

            self.treestore[iter_][Col.DETAIL.value] = signal.detail

    @Gtk.Template.Callback("on_user_data_edited")
    def __on_user_data_edited(self, renderer, path, new_text):
        iter_ = self.treestore.get_iter(path)
        signal = self.treestore[iter_][Col.SIGNAL.value]

        if signal is not None:
            if len(new_text) > 0:
                data_obj = self._object.project.get_object_by_name(signal.ui_id, new_text)

                if data_obj:
                    signal.user_data = data_obj.object_id
                    name = data_obj.name
                    signal.swap = True
                else:
                    signal.user_data = 0
                    signal.swap = False
                    self.treestore[iter_][Col.SWAP.value] = signal.swap
                    name = ""

                self.treestore[iter_][Col.USER_DATA.value] = name
            else:
                signal.user_data = 0
                signal.swap = False
                self.treestore[iter_][Col.USER_DATA.value] = ""

    @Gtk.Template.Callback("on_swap_toggled")
    def __on_swap_toggled(self, renderer, path):
        iter_ = self.treestore.get_iter(path)
        signal = self.treestore[iter_][Col.SIGNAL.value]

        if signal is not None:
            signal.swap = not self.treestore[iter_][Col.SWAP.value]
            self.treestore[iter_][Col.SWAP.value] = signal.swap

    @Gtk.Template.Callback("on_after_toggled")
    def __on_after_toggled(self, renderer, path):
        iter_ = self.treestore.get_iter(path)
        signal = self.treestore[iter_][Col.SIGNAL.value]

        if signal is not None:
            signal.after = not self.treestore[iter_][Col.AFTER.value]
            self.treestore[iter_][Col.AFTER.value] = signal.after

    def __on_signal_added(self, obj, signal):
        for row in self.treestore:
            if row[Col.OWNER_ID.value] == signal.owner_id:
                for child in row.iterchildren():
                    if child[Col.SIGNAL.value] is None and child[Col.SIGNAL_ID.value] == signal.signal_id:
                        self.treestore.insert_before(
                            row.iter,
                            child.iter,
                            (
                                signal,
                                signal.owner_id,
                                signal.signal_id,
                                signal.detail,
                                signal.handler,
                                str(signal.user_data),
                                signal.swap,
                                signal.after,
                                child[Col.INFO.value],
                                None,
                            ),
                        )
                break

    def __on_signal_removed(self, obj, signal):
        for row in self.treestore:
            for child in row.iterchildren():
                if child[Col.SIGNAL.value] == signal:
                    self.treestore.remove(child.iter)
                    return

    def __on_signal_changed(self, obj, signal):
        for row in self.treestore:
            for child in row.iterchildren():
                if child[Col.SIGNAL.value] == signal:
                    child[Col.DETAIL.value] = signal.detail
                    child[Col.HANDLER.value] = signal.handler
                    child[Col.USER_DATA.value] = str(signal.user_data)
                    child[Col.SWAP.value] = signal.swap
                    child[Col.AFTER.value] = signal.after
                    return

    def __populate_from_type(self, info, target):
        if len(info.signals) == 0:
            return None

        parent = self.treestore.append(None, (None, info.type_id, info.type_id, None, None, None, False, False, None, None))
        for signal_id in info.signals:
            signal = info.signals[signal_id]
            msg = utils.get_version_warning(target, signal.version, signal.deprecated_version, signal.signal_id)
            self.treestore.append(parent, (None, info.type_id, signal.signal_id, None, None, None, False, False, signal, msg))
        return parent

    def __populate_treestore(self):
        target = self._object.ui.get_target(self._object.info.library_id)

        # Populate object type signals
        parent = self.__populate_from_type(self._object.info, target)

        # Expand object type signals
        if parent:
            self.treeview.expand_row(self.treestore.get_path(parent), True)

        # Populate all hierarchy signals
        for type_id in self._object.info.hierarchy:
            info = self._object.project.type_info.get(type_id, None)
            if info:
                self.__populate_from_type(info, target)

        # Populate object signals
        for signal in self._object.signals:
            self.__on_signal_added(self._object, signal)

    def __signal_id_data_func(self, tree_column, cell, tree_model, iter_, column):
        info = tree_model[iter_][Col.INFO.value]
        signal_id = tree_model[iter_][Col.SIGNAL_ID.value]
        warning = tree_model[iter_][Col.WARNING_MESSAGE.value]

        if info and info.detailed:
            detail = tree_model[iter_][Col.DETAIL.value]
            signal = tree_model[iter_][Col.SIGNAL.value]

            cell.props.editable = False if signal is None else True
            cell.props.text = f"{signal_id}::{detail}" if detail is not None else signal_id
        else:
            cell.props.editable = False
            cell.props.text = signal_id

        cell.props.underline = Pango.Underline.ERROR if warning else Pango.Underline.NONE

    def __data_func(self, tree_column, cell, tree_model, iter_, column):
        info = tree_model[iter_][Col.INFO.value]
        signal = tree_model[iter_][Col.SIGNAL.value]

        if info is None:
            cell.props.visible = False
            return

        cell.props.visible = True

        if signal is None and column != Col.HANDLER.value:
            cell.props.sensitive = False
        else:
            cell.props.sensitive = True

        if signal and column == Col.USER_DATA.value:
            user_data = signal.user_data
            if user_data:
                data_obj = self._object.project.get_object_by_id(signal.ui_id, user_data)
                cell.props.text = data_obj.name if data_obj else ""
            else:
                cell.props.text = ""
